/*
///////////////////////////////////////////////////////////////////////////////////

    Partial STM32 library for GPIO demo for the STM32 Blue Pill
    Blog Part 2

    Copyright (C) 2021 - Martin Kaselis (modelrailmusings.weebly.com)

///////////////////////////////////////////////////////////////////////////////////
*/

#include "stm32.h"

#include <stdarg.h>

///////////////////////////////////////////////////////////////////////////////////////////

STM_FLASH *FLASH  = (STM_FLASH*) 0x40022000;
STM_RCC   *RCC    = (STM_RCC*)   0x40021000;
STM_STK   *STK    = (STM_STK*)   0xe000e010;
STM_GPIO  *GPIOA  = (STM_GPIO*)  0x40010800;
STM_GPIO  *GPIOB  = (STM_GPIO*)  0x40010c00;
STM_GPIO  *GPIOC  = (STM_GPIO*)  0x40011000;

///////////////////////////////////////////////////////////////////////////////////////////

// Temporary crash routine, will tie to serial output and assert() later on.
void crash( void )
{
    gpioInit();

    gpioPinInit( GPIO_PC13, GPIO_MODE_OUTPUT );

    for(;;)
    {
        gpioPinWrite( GPIO_PC13, 1 );
        delayUsec(50*1000);
        gpioPinWrite( GPIO_PC13, 0 );
        delayUsec(50*1000);
    }
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Initialise the system clock to the HSE at 72MHz, leave HSI running since some peripherals still need it.
void clockInit( void )
{
    // Conf clock : 72MHz using HSE 8MHz crystal w/ PLL X 9 (8MHz x 9 = 72MHz)
    FLASH->ACR = ( FLASH->ACR & ~FLASH_ACR_LATENCY_MASK ) | FLASH_ACR_LATENCY_2;      // Two wait states, per datasheet

    RCC->CFGR       |= RCC_CFGR_PPRE1_2;    // prescale AHB1 = HCLK/2
    RCC->CR         |= RCC_CR_HSEON;        // enable HSE clock
    while( !(RCC->CR & RCC_CR_HSERDY) )     // wait for the HSEREADY flag
        ;

    RCC->CFGR       |= RCC_CFGR_PLLSRC | RCC_CFGR_PLLMULL9; // set PLL source to HSE and multiply by 9
    RCC->CR         |= RCC_CR_PLLON;        // enable the PLL
    while( !(RCC->CR & RCC_CR_PLLRDY) )     // wait for the PLLRDY flag
        ;

    RCC->CFGR       |= RCC_CFGR_SW_PLL;     // set clock source to pll
    while( !(RCC->CFGR & RCC_CFGR_SWS_PLL) )   // wait for PLL to be CLK
        ;
}

// Delay N microseconds
void delayUsec( uint32_t usec )
{
    uint32_t i;

    // approximate calibration for 72MHz clock
    for( i = 0; i < 4 * usec; i++ )
        ;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Init systick subsystem
void systickInit( void )
{
    STK->RVR = 72000-1; // 1ms at 72MHz
    STK->CVR = 0;
    STK->CSR = STK_CSR_PROCESSOR_CLK | STK_CSR_INTERRUPT | STK_CSR_ENABLE;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Init the GPIO subsystem
void gpioInit(void)
{
    // RCC_APB2ENR: Enable I/O port C, B and A clocks, also AFIO clock for PB3/PB4 remapping (not used here)
    RCC->APB2ENR |= IOPC_EN | IOPB_EN | IOPA_EN | AFIOEN;
}

// Init one GPIO pin
void gpioPinInit( GPIO_BIT bit, GPIO_MODE mode )
{
    uint32_t *pReg;
    uint32_t shift = ( bit & 7 ) * 4;

    if(      bit <= GPIO_PA7  ) pReg = (uint32_t *) &GPIOA->CRL;
    else if( bit <= GPIO_PA15 ) pReg = (uint32_t *) &GPIOA->CRH;
    else if( bit <= GPIO_PB7  ) pReg = (uint32_t *) &GPIOB->CRL;
    else if( bit <= GPIO_PB15 ) pReg = (uint32_t *) &GPIOB->CRH;
    else if( bit <= GPIO_PC7  ) pReg = (uint32_t *) &GPIOC->CRL;
    else if( bit <= GPIO_PC15 ) pReg = (uint32_t *) &GPIOC->CRH;
    else
        crash();

    switch( mode )
    {
    case GPIO_MODE_INPUT:
        *pReg = ( *pReg & ~( 0x0f << shift ) ) | ( 0x04 << shift ); // input, floating (no pullup)
        break;
    case GPIO_MODE_OUTPUT:
        *pReg = ( *pReg & ~( 0x0f << shift ) ) | ( 0x01 << shift ); // output push-pull, max speed 10MHz
        break;
    case GPIO_MODE_ANALOG:
        *pReg = ( *pReg & ~( 0x0f << shift ) ) | ( 0x00 << shift ); // analog input
        break;
    case GPIO_MODE_AF_PUSHPULL:
        *pReg = ( *pReg & ~( 0x0f << shift ) ) | ( 0x09 << shift ); // alternate function push-pull output, 10 Mhz
        break;
    default:
        crash();
        break;
    }
}

// Write multiple bits under a mask.  Must all be in the same bank (A, B or C).  0 bits are cleared first, then 1 bits set.
void gpioPinWriteMultiple( GPIO_BIT bit, uint32_t mask, uint32_t value )
{
    uint32_t *pReg;
    uint32_t data;

    if(      bit <= GPIO_PA15 ) pReg = (uint32_t *) &GPIOA->BSRR;
    else if( bit <= GPIO_PB15 ) pReg = (uint32_t *) &GPIOB->BSRR;
    else if( bit <= GPIO_PC15 ) pReg = (uint32_t *) &GPIOC->BSRR;
    else
        crash();

    // turn 0 bits off, pause, then 1 bits on (protect FETs from H & L being on together)
    data = ( ( ~value & mask ) << 16 );
    *pReg = data;

    delayUsec( 1 );

    data = ( value & mask );
    *pReg = data;
}

// Write one GPIO pin
void gpioPinWrite( GPIO_BIT bit, uint32_t on )    // Takes approx 1usec per call
{
    uint32_t *pReg;
    uint32_t shift = ( bit & 15 );

    if(      bit <= GPIO_PA15 ) pReg = (uint32_t *) &GPIOA->BSRR;
    else if( bit <= GPIO_PB15 ) pReg = (uint32_t *) &GPIOB->BSRR;
    else if( bit <= GPIO_PC15 ) pReg = (uint32_t *) &GPIOC->BSRR;
    else
        crash();

    if( on )
        *pReg = 1 << shift;
    else
        *pReg = 1 << ( shift + 16 );
}

// Read one GPIO pin
uint32_t gpioPinRead( GPIO_BIT bit )
{
    uint32_t *pReg;
    uint32_t shift = ( bit & 15 );

    if(      bit <= GPIO_PA15 ) pReg = (uint32_t *) &GPIOA->IDR;
    else if( bit <= GPIO_PB15 ) pReg = (uint32_t *) &GPIOB->IDR;
    else if( bit <= GPIO_PC15 ) pReg = (uint32_t *) &GPIOC->IDR;
    else
        crash();

    return ( *pReg >> shift ) & 1;
}
